import click

from cli import client
from cli.gateway import GatewayManager
from cli.infra import InfraManager, cockpit
from cli.model import Route

DB_PASSWORD_OPTION = click.option(
    "--db-password",
    required=False,
    help="The password to use for the database. Will be generated if not provided.",
    envvar="DB_PASSWORD",
)
NO_METRICS_OPTION = click.option(
    "--no-metrics",
    is_flag=True,
    default=False,
    help="Disable Kong metrics integration with Cockpit.",
)


@click.group()
def cli():
    """CLI for managing the gateway.

    See the README for more information.
    """


@cli.command()
def local_config():
    """Sets up config for local deployment"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.set_up_config(True)


@cli.command()
def remote_config():
    """Sets up config for remote deployment"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.set_up_config(False)


@cli.command()
def get_routes():
    """Returns the routes configured on the gateway"""
    manager = GatewayManager()
    manager.print_routes()


@cli.command()
@click.argument("relative_url")
@click.argument("target")
@click.option(
    "--cors", is_flag=True, default=False, help="Add permissive cors to the route."
)
def add_route(relative_url, target, cors):
    """Adds a route to the gateway"""
    manager = GatewayManager()

    route = Route(relative_url, target, cors=cors)
    manager.add_route(route)


@cli.command()
@click.argument("relative_url")
@click.argument("target")
def delete_route(relative_url, target):
    """Deletes a route from the gateway"""
    manager = GatewayManager()

    route = Route(relative_url, target)
    manager.delete_route(route)


# TODO: integrate with existing CLI commands
@cli.command()
def setup_metrics():
    """Adds metrics plugin to the gateway"""
    gateway = GatewayManager()
    gateway.setup_global_kong_statsd_plugin()


@cli.command()
def create_admin_token():
    """Creates a token for the admin container"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    token = manager.create_admin_container_token()
    click.secho(token)


@cli.command()
def delete_containers():
    """Deletes the containers used for the gateway"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.delete_containers()


@cli.command()
def get_endpoint():
    """Returns the endpoint for the gateway"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    endpoint = manager.get_gateway_endpoint()
    click.secho(endpoint)


@cli.command()
def get_admin_endpoint():
    """Returns the endpoint for the gateway admin"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    endpoint = manager.get_gateway_admin_endpoint()
    click.secho(endpoint)


@cli.command()
@DB_PASSWORD_OPTION
@click.option(
    "--no-save", is_flag=True, default=False, help="Do not save the password."
)
def create_db(db_password: str | None, no_save: bool):
    """Creates the database for the gateway.

    If --no-save is passed, the password will not be saved to Secret Manager.
    The password will therefore need to be provided for other commands.
    """
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.create_db(password=db_password, save_password=not no_save)


@cli.command()
def delete_db():
    """Deletes the database for the gateway."""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.delete_db()


@cli.command()
def check_db():
    """Checks the status of the database"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.check_db()


@cli.command()
def await_db():
    """Waits for the database to be ready"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.await_db()


@cli.command()
def create_namespace():
    """Creates the container container namespace"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.create_namespace()


@cli.command()
def check_namespace():
    """Checks the status of the container namespace"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.check_namespace()


@cli.command()
def await_namespace():
    """Waits for the namespace to be ready"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.await_namespace()


@cli.command()
def delete_namespace():
    """Deletes the container namespace"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.delete_namespace()


@cli.command()
@DB_PASSWORD_OPTION
@NO_METRICS_OPTION
def create_containers(db_password: str | None, no_metrics: bool):
    """Creates the containers"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    if not no_metrics:
        # Check that Cockpit is activated
        no_metrics = not cockpit.ensure_cockpit_activated(scw_client=scw_client)
    manager.create_containers(db_password=db_password, forward_metrics=not no_metrics)


@cli.command()
def check_containers():
    """Checks the status of the containers"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.check_containers()


@cli.command()
def await_containers():
    """Waits for the containers to be ready"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.await_containers()


@cli.command()
def set_custom_domain():
    """Sets the custom domain for the gateway container"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)
    manager.set_custom_domain()


@cli.command()
@click.option(
    "--no-redeploy",
    is_flag=True,
    default=False,
    help="Don't redeploy the container, just update.",
)
@DB_PASSWORD_OPTION
def update_containers(
    no_redeploy: bool,
    db_password: str | None,
):
    """Updates the containers"""
    scw_client = client.get_scaleway_client()
    manager = InfraManager(scw_client)

    if no_redeploy:
        manager.update_container_without_deploy(db_password=db_password)
    else:
        manager.update_container(db_password=db_password)



from scaleway import Client
from scaleway_core.bridge import region

DEFAULT_API_REGION = region.REGION_FR_PAR


def get_scaleway_client() -> Client:
    """Create a Scaleway client."""
    scw_client = Client.from_config_file_and_env()
    if not scw_client.default_region:
        scw_client.default_region = DEFAULT_API_REGION
    return scw_client



import scaleway.rdb.v1 as sdk

DB_INSTANCE_NAME = "scw-sls-gw"
DB_ENGINE = "PostgreSQL-14"
DB_USERNAME = "kong"
DB_VOLUME_TYPE = "lssd"
DB_NODE_TYPE = "DB-DEV-S"
DB_VOLUME_SIZE = 5000000000  # Expressed in bytes


def create_database_instance(api: sdk.RdbV1API, password: str) -> sdk.Instance:
    """Create a managed database instance."""

    return api.create_instance(
        name=DB_INSTANCE_NAME,
        engine=DB_ENGINE,
        user_name=DB_USERNAME,
        password=password,
        is_ha_cluster=False,
        disable_backup=True,
        backup_same_region=True,
        node_type=DB_NODE_TYPE,
        volume_type=sdk.VolumeType(DB_VOLUME_TYPE),
        volume_size=DB_VOLUME_SIZE,
    )


def get_database_instance_by_name(
    api: sdk.RdbV1API, instance_name: str
) -> sdk.Instance | None:
    """Get a database instance by its name."""

    instances = api.list_instances_all(
        name=instance_name,
    )
    if not instances:
        return None

    return instances[0]



from concurrent import futures

import click
import scaleway.cockpit.v1beta1 as cpt
import scaleway.container.v1beta1 as cnt
import scaleway.function.v1beta1 as fnc
import scaleway.rdb.v1 as rdb
import scaleway.secret.v1alpha1 as sec
from loguru import logger
from scaleway import Client, ScalewayException

from cli import conf, infra


class InfraManager:
    """Manager for the infrastructure of the gateway.

    .. note::
        This class is used by the CLI to manage the infrastructure of the gateway.

        References:
            `Scaleway API documentation`_
            `Scaleway Python SDK`_

        .. _Scaleway API documentation:_ https://www.scaleway.com/en/developers/api/
        .. _Scaleway Python SDK:_ https://github.com/scaleway/scaleway-sdk-python
    """

    def __init__(self, scw_client: Client):
        # Initialise SCW client
        self.scw_client = scw_client

        # Init Scaleway APIs
        self.containers = cnt.ContainerV1Beta1API(
            self.scw_client, bypass_validation=True
        )
        self.functions = fnc.FunctionV1Beta1API(self.scw_client, bypass_validation=True)
        self.rdb = rdb.RdbV1API(self.scw_client, bypass_validation=True)
        self.secrets = sec.SecretV1Alpha1API(self.scw_client, bypass_validation=True)
        self.cockpit = cpt.CockpitV1Beta1API(self.scw_client, bypass_validation=True)

    def set_up_config(self, is_local: bool) -> None:
        """Set up the configuration for the gateway.

        Used by the GatewayManager to set up the route configuration.
        """
        if is_local:
            config = conf.InfraConfiguration.from_local()
        else:
            config = conf.InfraConfiguration.from_infra(self)
        config.save()

    def _get_namespace_or_abort(self) -> cnt.Namespace:
        namespace_name = infra.cnt.CONTAINER_NAMESPACE
        namespace = infra.cnt.get_namespace_by_name(self.containers, namespace_name)
        if not namespace:
            click.secho(
                f"Namespace {namespace_name} not found",
                fg="red",
                bold=True,
            )
            raise click.Abort()

        return namespace

    def _get_container_or_abort(self, admin: bool = False) -> cnt.Container:
        namespace = self._get_namespace_or_abort()
        container_name = (
            infra.cnt.CONTAINER_ADMIN_NAME if admin else infra.cnt.CONTAINER_NAME
        )
        container = infra.cnt.get_container_by_name(
            self.containers, namespace.id, container_name
        )

        if not container:
            container_name = "Admin" if admin else "Gateway"
            click.secho(
                f"{container_name} container not found.",
                fg="red",
                bold=True,
            )
            raise click.Abort()

        return container

    def _get_admin_container_or_abort(self) -> cnt.Container:
        return self._get_container_or_abort(admin=True)

    def _get_database_instance_or_abort(self) -> rdb.Instance:
        instance_name = infra.rdb.DB_INSTANCE_NAME
        instance = infra.rdb.get_database_instance_by_name(self.rdb, instance_name)
        if not instance:
            click.secho(
                f"Database instance {instance_name} not found",
                fg="red",
                bold=True,
            )
            raise click.Abort()

        return instance

    def _get_database_endpoint_or_abort(
        self, database_instance: rdb.Instance
    ) -> tuple[str, int]:
        endpoints = database_instance.endpoints
        if not endpoints:
            click.secho(
                f"Database instance {database_instance.name} has no endpoints",
                fg="red",
                bold=True,
            )
            raise click.Abort()

        endpoint = endpoints[0]
        address = endpoint.ip or endpoint.hostname
        if not address:
            click.secho(
                f"Database instance {database_instance.name} has no address",
                fg="red",
                bold=True,
            )
            raise click.Abort()

        return address, endpoint.port

    def _get_db_password_or_abort(self) -> str:
        try:
            password = infra.secrets.get_db_password(self.secrets)
            return password
        except ScalewayException as exception:
            if exception.status_code == 404:
                click.secho("Database password not found in Secret Manager", fg="red")
                raise click.Abort()
            raise exception

    def get_gateway_endpoint(self) -> str:
        """Get the endpoint of the gateway."""
        container = self._get_container_or_abort()
        return container.domain_name

    def get_gateway_admin_endpoint(self) -> str:
        """Get the endpoint of the gateway admin."""
        container = self._get_admin_container_or_abort()
        return container.domain_name

    def create_db(self, password: str | None, save_password: bool) -> None:
        """Create the database instance."""

        instance_name = infra.rdb.DB_INSTANCE_NAME
        instance = infra.rdb.get_database_instance_by_name(self.rdb, instance_name)
        if instance:
            click.secho(f"Database {instance_name} already exists")
            return

        click.secho(f"Creating database instance {instance_name}...")

        if not password:
            logger.info("Generating database password")
            password = infra.secrets.generate_database_password()

        if save_password:
            logger.info("Saving database password")
            infra.secrets.create_db_password_secret(self.secrets, password)

        infra.rdb.create_database_instance(self.rdb, password)
        click.secho("Database created", fg="green", bold="true")

    def check_db(self):
        """Check the status of the database instance."""
        instance = self._get_database_instance_or_abort()
        click.secho(f"Database status: {instance.status}", bold=True)

    def await_db(self) -> None:
        """Wait for the database instance to be ready."""
        instance = self._get_database_instance_or_abort()
        instance = self.rdb.wait_for_instance(instance_id=instance.id)

        if instance.status != rdb.InstanceStatus.READY:
            click.secho("Database is not ready", fg="red", bold="true")
            raise click.Abort()

        click.secho("Database ready", fg="green", bold=True)

    def delete_db(self) -> None:
        """Delete the database instance."""
        # Delete the secret
        infra.secrets.delete_db_password_secret(self.secrets)

        # Delete the database
        instance = self._get_database_instance_or_abort()
        self.rdb.delete_instance(instance_id=instance.id)
        click.secho("Database deleted", fg="green", bold=True)

    def create_namespace(self):
        """Create the namespace for the gateway."""
        namespace_name = infra.cnt.CONTAINER_NAMESPACE
        namespace = infra.cnt.get_namespace_by_name(self.containers, namespace_name)

        if namespace:
            click.secho("Namespace already exists")
            return

        click.secho(f"Creating namespace {namespace_name}...")
        infra.cnt.create_namespace(self.containers)

    def check_namespace(self):
        """Check the status of the namespace."""
        namespace = self._get_namespace_or_abort()
        if namespace.status == cnt.NamespaceStatus.ERROR:
            click.secho(
                f"Namespace in error: {namespace.error_message}",
                fg="red",
                bold="true",
            )
            raise click.Abort()

        click.secho(f"Namespace status: {namespace.status}", bold=True)

    def await_namespace(self):
        """Wait for the namespace to be ready."""
        namespace = self._get_namespace_or_abort()
        namespace = self.containers.wait_for_namespace(namespace_id=namespace.id)
        if namespace.status == cnt.NamespaceStatus.ERROR:
            click.secho(
                f"Namespace in error: {namespace.error_message}",
                fg="red",
                bold=True,
            )
            raise click.Abort()

        if namespace.status != cnt.NamespaceStatus.READY:
            click.secho("Namespace is not ready", fg="red", bold=True)
            raise click.Abort()

        click.secho("Namespace ready", fg="green", bold=True)

    def delete_namespace(self):
        """Delete the namespace."""
        namespace = self._get_namespace_or_abort()
        self.containers.delete_namespace(namespace_id=namespace.id)
        click.secho("Namespace deleted", fg="green", bold=True)

    def create_containers(self, db_password: str | None, forward_metrics: bool) -> None:
        """Create containers for Kong and Kong Admin."""
        database_instance = self._get_database_instance_or_abort()
        if not db_password:
            db_password = self._get_db_password_or_abort()
        db_host, db_port = self._get_database_endpoint_or_abort(database_instance)

        # Namespace should be created before creating containers
        namespace = self._get_namespace_or_abort()

        admin_container_name = infra.cnt.CONTAINER_ADMIN_NAME
        admin_container = infra.cnt.get_container_by_name(
            self.containers, namespace.id, admin_container_name
        )
        if admin_container:
            click.secho(
                f"Admin container {admin_container_name} already exists",
            )
        else:
            click.secho(
                f"Creating admin container {admin_container_name}",
            )
            created_container = infra.cnt.create_kong_admin_container(
                self.containers, namespace.id, db_host, db_port, db_password
            )
            click.echo(f"Deploying container {admin_container_name}")
            self.containers.deploy_container(container_id=created_container.id)

        container_name = infra.cnt.CONTAINER_NAME
        container = infra.cnt.get_container_by_name(
            self.containers, namespace.id, container_name
        )
        if container:
            click.secho(f"Container {container_name} already exists")
            return

        click.secho(f"Creating container {container_name}")

        token, metrics_push_url = None, None
        if forward_metrics:
            # Check if token exists
            token = infra.cpt.get_metrics_token(self.cockpit)
            if token:
                click.echo("Cockpit token already exists, recreating")
                infra.cpt.delete_metrics_token(self.cockpit, token)
            else:
                click.echo("Creating Cockpit token")

            token_key = infra.cpt.create_metrics_token(self.cockpit)
            metrics_push_url = infra.cpt.get_metrics_push_url(self.cockpit)

        created_container = infra.cnt.create_kong_container(
            self.containers,
            namespace.id,
            db_host,
            db_port,
            db_password,
            metrics_token=token_key,
            metrics_push_url=metrics_push_url,
        )

        click.secho(f"Deploying container {container_name}")
        self.containers.deploy_container(container_id=created_container.id)

    def check_containers(self):
        """Check the status of the containers."""
        admin_container = self._get_admin_container_or_abort()
        container = self._get_container_or_abort()

        click.secho(f"Admin container status: {admin_container.status}", bold=True)
        click.secho(f"Container status: {container.status}", bold=True)

    def await_containers(self):
        """Wait for the containers to be ready."""
        admin_container = self._get_admin_container_or_abort()
        container = self._get_container_or_abort()

        # Execute in parallel
        results = futures.ThreadPoolExecutor(max_workers=2).map(
            lambda id: self.containers.wait_for_container(container_id=id),
            (admin_container.id, container.id),
        )
        for res in results:
            if res.status == cnt.ContainerStatus.ERROR:
                click.secho(
                    f"Container {res.name} is in error: {res.error_message}"
                    "Check the logs for more details.",
                    fg="red",
                    bold=True,
                )
                raise click.Abort()
            if res.status != cnt.ContainerStatus.READY:
                click.secho(f"Container {res.name} is not ready", fg="red", bold=True)
                raise click.Abort()
        click.secho("Containers are ready", fg="green", bold=True)

    def delete_containers(self):
        """Delete the containers."""
        admin_container = self._get_admin_container_or_abort()
        container = self._get_container_or_abort()

        click.secho(f"Deleting container {admin_container.name}")
        self.containers.delete_container(container_id=admin_container.id)

        click.secho(f"Deleting container {container.name}")
        self.containers.delete_container(container_id=container.id)

    def get_function_endpoint(
        self, namespace_name: str, function_name: str
    ) -> str | None:
        """Get the endpoint of a function."""
        return infra.fnc.get_function_endpoint_by_name(
            self.functions, namespace_name, function_name
        )

    def create_admin_container_token(self) -> str:
        """Create a token to access the private admin container."""
        admin_container = self._get_admin_container_or_abort()
        token = self.containers.create_token(container_id=admin_container.id)
        return token.token

    def update_container(self, db_password: str | None):
        """Update the container."""
        self.update_container_without_deploy(db_password=db_password)

        admin_container = self._get_admin_container_or_abort()
        container = self._get_container_or_abort()

        click.secho("Deploying admin container...")
        self.containers.deploy_container(container_id=admin_container.id)

        click.secho("Deploying container...")
        self.containers.deploy_container(container_id=container.id)

    def update_container_without_deploy(self, db_password: str | None):
        """Update the container without deploying it."""
        admin_container = self._get_admin_container_or_abort()
        container = self._get_container_or_abort()

        database_instance = self._get_database_instance_or_abort()
        db_host, db_port = self._get_database_endpoint_or_abort(database_instance)
        if not db_password:
            db_password = self._get_db_password_or_abort()

        click.echo(f"Updating container {admin_container.name}")
        infra.cnt.update_kong_admin_container(
            self.containers, admin_container.id, db_host, db_port, db_password
        )
        click.echo(f"Updating container {container.name}")

        token, metrics_push_url = None, None
        if container.environment_variables.get("FORWARD_METRICS"):
            token = infra.cpt.get_metrics_token(self.cockpit)
            if token:
                infra.cpt.delete_metrics_token(self.cockpit, token)

            click.echo("Creating Cockpit token to forward metrics...")
            token_key = infra.cpt.create_metrics_token(self.cockpit)
            metrics_push_url = infra.cpt.get_metrics_push_url(self.cockpit)

        infra.cnt.update_kong_container(
            self.containers,
            container.id,
            db_host,
            db_port,
            db_password,
            metrics_token=token_key,
            metrics_push_url=metrics_push_url,
        )

    def set_custom_domain(self):
        """Set a custom domain for the container."""
        raise NotImplementedError



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Route(object):
    relative_url: str
    target: str

    http_methods: Optional[List[str]] = None
    cors: Optional[bool] = False

    @property
    def name(self):
        return self.relative_url.replace("/", "_")

    def route_json(self):
        return {
            "name": self.name,
            "paths": [self.relative_url],
            "service": {"name": self.name},
            "methods": self.http_methods,
        }

    def service_json(self):
        return {"name": self.name, "url": self.target}

    def cors_json(self):
        return {
            "name": "cors",
            "config": {"origins": ["*"], "headers": ["*"], "credentials": True},
        }

    def __eq__(self, other):
        equal = True
        equal &= self.relative_url == other.relative_url
        equal &= self.target == other.target
        equal &= self.http_methods == other.http_methods
        equal &= self.cors == other.cors

        return equal



import click
import scaleway.cockpit.v1beta1 as sdk
from scaleway import Client, ScalewayException

METRICS_TOKEN_NAME = "scw-gw-write-metrics"
WRITE_METRICS_SCOPE = sdk.TokenScopes(
    query_metrics=False,
    write_metrics=True,
    setup_metrics_rules=False,
    query_logs=False,
    write_logs=False,
    setup_logs_rules=False,
    setup_alerts=False,
)


def ensure_cockpit_activated(scw_client: Client) -> bool:
    """Ensure Cockpit is activated.

    Prompt the user to activate cockpit if it is not activated.
    """
    api = sdk.CockpitV1Beta1API(scw_client)
    try:
        api.get_cockpit()
        return True
    except ScalewayException as err:
        if not err.status_code == 404:
            raise

        should_activate_cockpit = click.confirm(
            "Cockpit not activated for project. Do you want to activate it?",
        )

        if not should_activate_cockpit:
            return False
        cockpit = api.activate_cockpit()
        api.wait_for_cockpit(project_id=cockpit.project_id)
        click.secho("Cockpit activated", fg="green")

        return True


def get_metrics_push_url(api: sdk.CockpitV1Beta1API) -> str:
    """Get the Cockpit metrics push URL."""
    cockpit = api.get_cockpit()

    if not cockpit.endpoints:
        # Should never happen
        raise RuntimeError("Cockpit has no endpoints")

    return cockpit.endpoints.metrics_url + "/api/v1/push"


def get_metrics_token(api: sdk.CockpitV1Beta1API) -> sdk.Token | None:
    """Get the Cockpit token used to write metrics."""
    tokens = api.list_tokens_all()

    for token in tokens:
        if token.name == METRICS_TOKEN_NAME:
            return token

    return None


def create_metrics_token(api: sdk.CockpitV1Beta1API) -> str:
    """Create a Cockpit token to write metrics."""
    token = api.create_token(
        name=METRICS_TOKEN_NAME,
        scopes=WRITE_METRICS_SCOPE,
    )

    if not token.secret_key:
        # Should never happen
        raise RuntimeError("Token has no secret key")

    return token.secret_key


def delete_metrics_token(api: sdk.CockpitV1Beta1API, token: sdk.Token) -> None:
    """Delete a Cockpit token."""
    api.delete_token(token_id=token.id)



import scaleway.container.v1beta1 as sdk

from cli.conf import DB_DATABASE_NAME
from cli.infra.image import IMAGE_TAG
from cli.infra.rdb import DB_USERNAME

CONTAINER_NAMESPACE = "scw-sls-gw"

CONTAINER_NAME = "scw-sls-gw"
CONTAINER_MIN_SCALE = 1
CONTAINER_MAX_SCALE = 5
CONTAINER_MEMORY_LIMIT = 1024

CONTAINER_ADMIN_NAME = "scw-sls-gw-admin"
CONTAINER_ADMIN_MIN_SCALE = 1
CONTAINER_ADMIN_MAX_SCALE = 1
CONTAINER_ADMIN_MEMORY_LIMIT = 1024
CONTAINER_ADMIN_PORT = 8001


def create_namespace(api: sdk.ContainerV1Beta1API) -> sdk.Namespace:
    """Create a namespace for the containers."""
    return api.create_namespace(
        name=CONTAINER_NAMESPACE,
    )


def get_namespace_by_name(
    api: sdk.ContainerV1Beta1API, namespace_name: str
) -> sdk.Namespace | None:
    """Get a namespace by its name."""
    namespaces = api.list_namespaces_all(
        name=namespace_name,
    )
    if not namespaces:
        return None

    return namespaces[0]


def get_container_by_name(
    api: sdk.ContainerV1Beta1API, namespace_id: str, container_name: str
) -> sdk.Container | None:
    """Get a container by its name."""
    containers = api.list_containers_all(
        namespace_id=namespace_id,
        name=container_name,
    )
    if not containers:
        return None

    return containers[0]


def get_base_container_env_vars(db_host: str, db_port: int) -> dict[str, str]:
    """Get the base environment variables for the container."""
    return {
        "KONG_PG_HOST": db_host,
        "KONG_PG_PORT": str(db_port),
        "KONG_PG_DATABASE": DB_DATABASE_NAME,
        "KONG_PG_USER": DB_USERNAME,
    }


def get_base_secret_env_vars(db_password: str) -> list[sdk.Secret]:
    """Get the secret environment variables for the container."""
    return [
        sdk.Secret(key="KONG_PG_PASSWORD", value=db_password),
    ]


def create_kong_container(
    api: sdk.ContainerV1Beta1API,
    namespace_id: str,
    db_host: str,
    db_port: int,
    db_password: str,
    metrics_token: str | None,
    metrics_push_url: str | None,
) -> sdk.Container:
    """Create the Kong container."""
    env_vars = get_base_container_env_vars(db_host=db_host, db_port=db_port)
    secret_env_vars = get_base_secret_env_vars(db_password=db_password)

    if metrics_token and metrics_push_url:
        secret_env_vars.append(sdk.Secret("COCKPIT_METRICS_TOKEN", metrics_token))
        env_vars["FORWARD_METRICS"] = "1"
        env_vars["COCKPIT_METRICS_PUSH_URL"] = metrics_push_url

    return api.create_container(
        namespace_id=namespace_id,
        name=CONTAINER_NAME,
        memory_limit=CONTAINER_MEMORY_LIMIT,
        min_scale=CONTAINER_MIN_SCALE,
        max_scale=CONTAINER_MAX_SCALE,
        privacy=sdk.ContainerPrivacy.PUBLIC,
        protocol=sdk.ContainerProtocol.HTTP1,
        http_option=sdk.ContainerHttpOption.REDIRECTED,
        registry_image=IMAGE_TAG,
        environment_variables=env_vars,
        secret_environment_variables=secret_env_vars,
    )


def create_kong_admin_container(
    api: sdk.ContainerV1Beta1API,
    namespace_id: str,
    db_host: str,
    db_port: int,
    db_password: str,
) -> sdk.Container:
    """Create the Kong admin container."""
    env_vars = get_base_container_env_vars(db_host=db_host, db_port=db_port)
    env_vars["IS_ADMIN_CONTAINER"] = "1"

    secret_env_vars = get_base_secret_env_vars(db_password=db_password)

    return api.create_container(
        namespace_id=namespace_id,
        name=CONTAINER_ADMIN_NAME,
        memory_limit=CONTAINER_ADMIN_MEMORY_LIMIT,
        min_scale=CONTAINER_ADMIN_MIN_SCALE,
        max_scale=CONTAINER_ADMIN_MAX_SCALE,
        port=CONTAINER_ADMIN_PORT,
        privacy=sdk.ContainerPrivacy.PRIVATE,
        protocol=sdk.ContainerProtocol.HTTP1,
        http_option=sdk.ContainerHttpOption.REDIRECTED,
        registry_image=IMAGE_TAG,
        environment_variables=env_vars,
        secret_environment_variables=secret_env_vars,
    )


def update_kong_container(
    api: sdk.ContainerV1Beta1API,
    container_id: str,
    db_host: str,
    db_port: int,
    db_password: str,
    metrics_token: str | None,
    metrics_push_url: str | None,
) -> sdk.Container:
    """Create the Kong container."""
    env_vars = get_base_container_env_vars(db_host=db_host, db_port=db_port)
    secret_env_vars = get_base_secret_env_vars(db_password=db_password)

    if metrics_token and metrics_push_url:
        secret_env_vars.append(sdk.Secret("COCKPIT_METRICS_TOKEN", metrics_token))
        env_vars["FORWARD_METRICS"] = "1"
        env_vars["COCKPIT_METRICS_PUSH_URL"] = metrics_push_url

    return api.update_container(
        container_id=container_id,
        memory_limit=CONTAINER_MEMORY_LIMIT,
        min_scale=CONTAINER_MIN_SCALE,
        max_scale=CONTAINER_MAX_SCALE,
        privacy=sdk.ContainerPrivacy.PUBLIC,
        protocol=sdk.ContainerProtocol.HTTP1,
        http_option=sdk.ContainerHttpOption.REDIRECTED,
        registry_image=IMAGE_TAG,
        environment_variables=env_vars,
        secret_environment_variables=secret_env_vars,
    )


def update_kong_admin_container(
    api: sdk.ContainerV1Beta1API,
    container_id: str,
    db_host: str,
    db_port: int,
    db_password: str,
) -> sdk.Container:
    """Create the Kong admin container."""
    env_vars = get_base_container_env_vars(db_host=db_host, db_port=db_port)
    env_vars["IS_ADMIN_CONTAINER"] = "1"

    secret_env_vars = get_base_secret_env_vars(db_password=db_password)

    return api.update_container(
        container_id=container_id,
        memory_limit=CONTAINER_ADMIN_MEMORY_LIMIT,
        min_scale=CONTAINER_ADMIN_MIN_SCALE,
        max_scale=CONTAINER_ADMIN_MAX_SCALE,
        port=CONTAINER_ADMIN_PORT,
        privacy=sdk.ContainerPrivacy.PRIVATE,
        protocol=sdk.ContainerProtocol.HTTP1,
        http_option=sdk.ContainerHttpOption.REDIRECTED,
        registry_image=IMAGE_TAG,
        environment_variables=env_vars,
        secret_environment_variables=secret_env_vars,
    )



import os
import typing as t
from dataclasses import asdict, dataclass

import yaml

if t.TYPE_CHECKING:
    # Importing conditionally to avoid circular imports
    from cli.infra import InfraManager

USER_HOME = os.path.expanduser("~")
CONFIG_DIR = os.path.join(USER_HOME, ".config", "scw")
CONFIG_FILE = os.path.join(CONFIG_DIR, "gateway.yml")

# Name is fixed for all managed databases
DB_DATABASE_NAME = "rdb"
DB_DATABASE_NAME_LOCAL = "kong"


@dataclass
class InfraConfiguration:
    """Configuration used to store information of a deployed gateway."""

    protocol: t.Literal["http", "https"]
    gw_admin_host: str
    gw_admin_port: str
    gw_admin_token: str
    gw_host: str
    gw_port: str
    db_host: str
    db_port: str
    db_name: str

    @staticmethod
    def from_local() -> "InfraConfiguration":
        """Get the configuration for the local docker-compose stack."""
        return InfraConfiguration(
            protocol="http",
            gw_admin_host="localhost",
            gw_admin_port="8001",
            gw_admin_token="",
            gw_host="localhost",
            gw_port="8080",
            db_host="localhost",
            db_port="5432",
            db_name=DB_DATABASE_NAME_LOCAL,
        )

    @staticmethod
    def from_infra(manager: "InfraManager") -> "InfraConfiguration":
        admin_host = manager.get_gateway_admin_endpoint()
        container_host = manager.get_gateway_endpoint()

        instance = manager._get_database_instance_or_abort()
        endpoint = instance.endpoints[0]

        token = manager.create_admin_container_token()

        return InfraConfiguration(
            protocol="https",
            gw_admin_host=admin_host,
            gw_admin_port="",
            gw_admin_token=token,
            gw_host=container_host,
            gw_port="",
            db_host=str(endpoint.ip),
            db_port=str(endpoint.port),
            db_name=DB_DATABASE_NAME,
        )

    @staticmethod
    def load() -> "InfraConfiguration":
        """Read the configuration from the config file."""
        with open(CONFIG_FILE, mode="rt", encoding="utf-8") as file:
            conf = yaml.safe_load(file)
            return InfraConfiguration(**conf)

    def save(self) -> None:
        """Save the configuration to a file."""
        os.makedirs(CONFIG_DIR, exist_ok=True)

        with open(CONFIG_FILE, mode="wt", encoding="utf-8") as file:
            yaml.safe_dump(asdict(self), file)



import json
from typing import List

import click
import requests
from loguru import logger

from cli import conf
from cli.model import Route


def response_hook(response: requests.Response, *_args, **_kwargs):
    """Response hook to log request and call raise_for_status."""
    req = response.request
    logger.debug(f"{req.method}: {req.url}")
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        body_json = err.response.json()
        logger.error(f"Error: \n{json.dumps(body_json, indent=2)}")
        raise


class GatewayManager:
    """Configure routes via the Kong admin API."""

    admin_url: str
    gateway_url: str

    def __init__(self):
        # Local local config
        self.config = conf.InfraConfiguration.load()

        admin_url = [
            self.config.protocol,
            "://",
            self.config.gw_admin_host,
        ]

        admin_port = self.config.gw_admin_port
        if admin_port:
            admin_url.extend([":", str(admin_port)])

        self.admin_url = "".join(admin_url)

        gateway_url = [
            self.config.protocol,
            "://",
            self.config.gw_host,
        ]

        gateway_port = self.config.gw_port
        if gateway_port:
            gateway_url.extend([":", str(gateway_port)])

        self.gateway_url = "".join(gateway_url)

        self.routes_url = self.admin_url + "/routes"
        self.services_url = self.admin_url + "/services"
        self.token = self.config.gw_admin_token

        self.session = requests.Session()
        if self.token:
            self.session.headers["X-Auth-Token"] = self.token
        self.session.headers["Content-Type"] = "application/json"
        self.session.hooks["response"].append(response_hook)

    def add_route(self, route: Route):
        service_url = f"{self.services_url}/{route.name}"
        route_url = f"{self.routes_url}/{route.name}"

        resp = self.session.put(url=service_url, json=route.service_json())
        resp = self.session.put(url=route_url, json=route.route_json())

        plugins_url = f"{route_url}/plugins"
        if route.cors:
            try:
                self.session.post(url=plugins_url, json=route.cors_json())
            except requests.HTTPError:
                # TODO - avoid ignoring this, any way to create or update?
                pass

        return resp

    def delete_route(self, route: Route):
        self.session.delete(url=f"{self.routes_url}/{route.name}")
        resp = self.session.delete(url=f"{self.services_url}/{route.name}")
        return resp

    def print_routes(self):
        routes: List[Route] = self.get_routes()
        routes.sort(key=lambda r: r.relative_url)

        click.secho(
            f"{'RELATIVE URL':<25} {'TARGET':<40} {'HTTP_METHODS':<10}", bold=True
        )
        for r in routes:
            http_methods = r.http_methods if r.http_methods else "All"
            click.secho(f"{r.relative_url:<25} {r.target:<40} {http_methods:<10}")

    def get_routes(self) -> List[Route]:
        resp = self.session.get(url=self.routes_url)
        route_data = {r.get("name"): r for r in resp.json().get("data")}

        resp = self.session.get(url=self.services_url)
        service_data = {s.get("name"): s for s in resp.json().get("data")}

        routes = list()
        for _, r in route_data.items():
            route_name = r["name"]
            route_path = r["paths"][0]
            http_methods = r.get("methods")

            service = service_data.get(route_name)
            if not service:
                continue

            service_port = service["port"]
            service_host = service["host"]
            service_protocol = service["protocol"]
            service_url = f"{service_protocol}://{service_host}:{service_port}"

            routes.append(
                Route(
                    relative_url=route_path,
                    target=service_url,
                    http_methods=http_methods,
                )
            )

        return routes

    def setup_global_kong_statsd_plugin(self) -> str:
        """Install the kong statsd plugin on the kong admin API.

        This plugin is used to send metrics to Cockpit.
        It is installed globally for all services.
        """
        plugins_url = self.admin_url + "/plugins"

        # Delete existing statsd plugin if it exists
        resp = self.session.get(plugins_url)
        plugins = resp.json()["data"]
        for plugin in plugins:
            if plugin["name"] == "statsd":
                self.session.delete(f"{plugins_url}/{plugin['id']}")

        # Install statsd plugin
        resp = self.session.post(
            plugins_url,
            json={
                "name": "statsd",
                "config": {
                    "port": 8125,
                    "prefix": "kong",
                },
            },
        )
        body_json = resp.json()
        plugin_id = body_json["id"]
        return plugin_id



from importlib.metadata import version

IMAGE_REGISTRY = "docker.io"
IMAGE_ORG = "scaleway"
IMAGE_NAME = "serverless-gateway"

# Single source of truth for the image version
# is the version of the package
IMAGE_VERSION = version("scw-gateway")

IMAGE_TAG = f"{IMAGE_REGISTRY}/{IMAGE_ORG}/{IMAGE_NAME}:{IMAGE_VERSION}"



# pylint: disable=useless-import-alias
# flake8: noqa
from . import cockpit as cpt
from . import container as cnt
from . import function as fnc
from . import image as image
from . import rdb as rdb
from . import secrets as secrets
from .manager import InfraManager as InfraManager



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cli', 'cli.infra']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'loguru==0.6.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.2,<3.0.0',
 'scaleway>=0.12.0,<0.13.0']

entry_points = \
{'console_scripts': ['scwgw = cli.cli:cli']}

setup_kwargs = {
    'name': 'scw-gateway',
    'version': '0.2.0',
    'description': 'CLI to deploy and manage a self-hosted Kong gateway on Scaleway Serverless Ecosystem',
    'long_description': '',
    'author': 'Simon Shillaker',
    'author_email': 'sshillaker@scaleway.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import scaleway.function.v1beta1 as sdk


def get_function_endpoint_by_name(
    api: sdk.FunctionV1Beta1API, namespace_name: str, function_name: str
) -> str | None:
    """Get the endpoint of a function by its name."""

    namespaces = api.list_namespaces_all(
        name=namespace_name,
    )
    if not namespaces:
        return None

    namespace = namespaces[0]
    functions = api.list_functions_all(namespace_id=namespace.id, name=function_name)

    return functions[0].domain_name if functions else None
